<?php

namespace Drupal\couchbasedrupal;

/**
 * We need some extra magic around
 * the bucket.
 */
class CouchbaseBucket {

  /**
   * Couchbase bucket
   *
   * @var \CouchbaseBucket
   */
  protected $bucket;

  /**
   * Bucket name.
   *
   * @var string
   */
  protected $name;

  /**
   * If N1ql client is enabled.
   *
   * @var bool
   */
  protected $N1qlEnabled = FALSE;

  /**
   * Fix expiration for better precision.
   *
   * @see http://docs.couchbase.com/developer/dev-guide-3.0/doc-expiration.html
   */
  public function FixExpiration($expire) {
    // Because time() might not be exactly the same in PHP and in Couchbase,
    // use reltive TTL() instead of absolute for small timeouts.
    $timeout = $expire - time();
    if ($timeout < 30 * 24 * 60 * 60) {
      $result = (int) $timeout;
    }
    else {
      $result = (int) $expire;
    }

    return $result > 0 ? $result : 1;
  }

  /**
   * Get an instance of CouchbaseBucket
   *
   * @param \CouchbaseCluster $cluster
   * @param mixed $name
   * @param TranscoderInterface $transcoder
   */
  public function __construct(\CouchbaseCluster $cluster, $name, TranscoderInterface $transcoder = NULL) {
    $this->bucket = $cluster->openBucket($name);
    // If the user provides a transcoder, use it.
    if ($transcoder) {
      $this->bucket->setTranscoder([$transcoder, 'encode'], [$transcoder, 'decode']);
    }
    $this->name = $name;
  }

  /**
   * Retrieve an item from the bucket. Returns
   * FALSE if it does not exist.
   *
   * @param string $key
   * @return mixed|bool
   */
  public function get($key) {
    try {
      return $this->bucket->get($key);
    }
    catch (\Exception $e) {
      if ($e->getCode() == 13) {
        return FALSE;
      }
      throw $e;
    }
  }

  /**
   * Retrieve a list of items from the bucket.
   *
   * @param string[] $key
   * @return array
   */
  public function getMultiple($keys) {
    return $this->bucket->get($keys);
  }

  /**
   * Remove an item from the bucket without
   * throwing and exceptino if it does not exist.
   *
   * @param string $keys
   *   A key or an array of keys to remove.
   * @return bool
   *   Always returns TRUE.
   */
  public function remove($key) {
    try {
      // We can delete all of them at once!
      $this->bucket->remove($key);
    }
    catch (\Exception $e) {
      if ($e->getCode() == 13) {
        return FALSE;
      }
      throw $e;
    }
    return TRUE;
  }

  /**
   * Remove an item from the bucket without
   * throwing and exceptino if it does not exist.
   *
   * @param string[] $keys
   *   An array of keys to remove.
   * @return bool
   *   Always returns TRUE.
   */
  public function removeMultiple($keys) {
    // We can delete all of them at once!
    $this->bucket->remove($keys);
    return TRUE;
  }

  /**
   * Upsert an intem in the bucket.
   *
   * @param string $id
   * @param mixed $val
   * @param array $options
   */
  public function upsert($id, $val = NULL, $options = array()) {
    $this->bucket->upsert($id, $val, $options);
  }

  /**
   * Insert an item in a bucket. If the item
   * already exists returns FALSE.
   *
   * @param string $id
   * @param mixed $val
   * @param mixed $options
   * @return bool
   */
  public function insert($id, $val = NULL, $options = array()) {
    try {
      $this->bucket->insert($id, $val, $options);
      return TRUE;
    }
    catch (\Exception $e) {
      $code = $e->getCode();
      // 12,13 are document already exists or document does not exist.
      if (in_array($code, array(12, 13))) {
        return FALSE;
      }
      throw $e;
    }
  }

  /**
   * Summary of touch
   *
   * @param string $id
   * @param int $expiry
   * @param array $options
   * @return mixed
   */
  public function touch($id, $expiry, array $options = []) {
    $result = $this->bucket->touch($id, $expiry, $options);
    return $result;
  }

  /**
   * Summary of counter
   *
   * @param mixed $ids
   * @param mixed $ids
   * @param array $options
   * @return mixed
   */
  public function counter($ids, $delta, array $options = []) {
    $result = $this->bucket->counter($ids, $delta, $options);
    return $result;
  }

  /**
   * Get the bucket manager.
   *
   * @return \CouchbaseBucketManager
   */
  public function manager() {
    return $this->bucket->manager();
  }

  /**
   * Delete elements by a prefix.
   *
   * @param string $prefix
   */
  public function deleteAllByPrefix($prefix) {
    $result = [];
    // TODO: Escape the prefix
    // @see http://developer.couchbase.com/documentation/server/4.0/n1ql/n1ql-language-reference/literals.html
    $query = "DELETE FROM {$this->name} WHERE META({$this->name}).id LIKE \"{$prefix}%\"";
    $result = $this->queryN1QL(\CouchbaseN1qlQuery::fromString($query), NULL, TRUE);
    return array_combine(array_column($result, 'id'), array_column($result, 'cid'));
  }

  public function getName() {
    return $this->name;
  }

  /**
   * Return all keys that start with a prefix
   * using N1QL.
   *
   * @param string $prefix
   * @return mixed
   */
  public function getAllKeysByPrefix($prefix) {
    $prefix_length = strlen($prefix);
    $result = [];
    // TODO: Escape the prefix
    // @see http://developer.couchbase.com/documentation/server/4.0/n1ql/n1ql-language-reference/literals.html
    $query = "SELECT SUBSTR(META({$this->name}).id, $prefix_length) AS cid, META({$this->name}).id as id FROM {$this->name} WHERE META({$this->name}).id LIKE \"{$prefix}%\"";
    $result = $this->queryN1QL(\CouchbaseN1qlQuery::fromString($query), NULL, TRUE);
    return array_combine(array_column($result, 'id'), array_column($result, 'cid'));
  }

  /**
   * Return all keys that start with a prefix
   * using N1QL.
   *
   * Implementation should not be based on document contents
   * because it is shared between Backend and RawBackend.
   *
   * @param string $prefix
   * @return mixed
   */
  public function getAllItemsByPrefix($prefix) {

    // Due to limitations in N1QL the most easy implementation
    // right now is to retrieve the keys and then the elements
    // because N1QL can only retrieve full documents and not
    // binary values.
    $keys = $this->getAllKeysByPrefix($prefix);
    return $this->getMultiple(array_keys($keys));

    //$prefix_length = strlen($prefix);
    //$result = [];
    //// TODO: Escape the prefix
    //// @see http://developer.couchbase.com/documentation/server/4.0/n1ql/n1ql-language-reference/literals.html
    //$query = "SELECT SUBSTR(META({$this->name}).id, $prefix_length) AS cid, META({$this->name}).id as id, TYPE(default) as type, {$this->name} as data FROM {$this->name} WHERE META({$this->name}).id LIKE \"{$prefix}%\"";
    //try {
    //  $result = $this->queryN1QL(\CouchbaseN1qlQuery::fromString($query));
    //}
    //catch (\Exception $e) {
    //  // CREATE INDEX idx ON mybucket(`key`, `value` IS VALUED) WHERE `value` IS VALUED;
    //  // https://forums.couchbase.com/t/explanation-on-primary-scan-secondary-scan/6048/12
    //  if ($e->qCode == 4000) {
    //    $query = \CouchbaseN1qlQuery::fromString("CREATE PRIMARY INDEX ON {$this->name} USING GSI");
    //    $this->queryN1QL($query);
    //    $result = $this->queryN1QL(\CouchbaseN1qlQuery::fromString($query));
    //  }
    //  throw $e;
    //}
    //// N1SQL does not allow to retrieve directly binary items, so we need an
    //// extra roundtrip to retrieve them.
    //// @see https://issues.couchbase.com/browse/DOC-704
    //$items = [];
    //$missing = [];
    //foreach ($result as $data) {
    //  if ($data->type == 'binary') {
    //    $missing[] = $data->id;
    //  }
    //  else {
    //    $items[$data->cid] = $data->data;
    //  }
    //}
    //$missing = $this->getMultiple($missing);
    //return $items;
  }

  /**
   * Enables client side N1QL
   */
  public function enableN1ql() {
    if (!$this->N1qlEnabled) {
      $this->bucket->enableN1ql(array('http://1.1.1.1:8093/','http://1.1.1.2:8093/'));
      $this->N1qlEnabled = TRUE;
    }
  }

  /**
   * Run a N1QL query
   *
   * @param string $query
   */
  public function queryN1QL(\CouchbaseN1qlQuery $query, $params = NULL, $json_array = FALSE) {
    $this->enableN1ql();
    try {
      return $this->bucket->query($query, $params, $json_array);
    }
    catch (\Exception $e) {
      // CREATE INDEX idx ON mybucket(`key`, `value` IS VALUED) WHERE `value` IS VALUED;
      // https://forums.couchbase.com/t/explanation-on-primary-scan-secondary-scan/6048/12
      if ($e->qCode == 4000) {
        $query = \CouchbaseN1qlQuery::fromString("CREATE PRIMARY INDEX ON {$this->name} USING GSI");
        $this->queryN1QL($query);
        return $this->queryN1QL(\CouchbaseN1qlQuery::fromString($query));
      }
      throw $e;
    }
  }

  /**
   * Run a view query
   *
   * @param \CouchbaseN1qlQuery $query
   * @param mixed $params
   * @param mixed $json_array
   * @return mixed
   */
  public function queryView(\CouchbaseViewQuery $query, $params = NULL, $json_array = FALSE) {
    if ($params == NULL && $json_array == NULL) {
      return $this->bucket->query($query);
    }
    else if ($json_array != FALSE) {
      return $this->bucket->query($query, $params, $json_array);
    }
    else {
      return $this->bucket->query($query, $params);
    }
  }

  /**
   * Clear all the documents (views) in this
   * bucket.
   */
  public function clearAllDocuments() {
    $manager = $this->manager();
    $documents = $manager->getDesignDocuments();
    foreach ($documents as $name => $definition) {
      $manager->removeDesignDocument($name);
    }
  }

  /**
   * Get information about an index, or false
   * if it does not exist.
   * 
   * @param string $name 
   * @return mixed
   */
  public function indexGet($name) {
    $result = $this->queryN1QL(\CouchbaseN1qlQuery::fromString("SELECT * FROM system:indexes WHERE name='$name'", NULL, FALSE));
    if (isset($result[0]['indexes'])) {
      return $result[0]['indexes'];
    }
    return FALSE;
  }

  public function indexDrop($name) {
    $result = $this->queryN1QL(\CouchbaseN1qlQuery::fromString("DROP INDEX $name", NULL, FALSE));
  }
}
